@@ -1,6 +1,6 @@
 /* ncftpput.c
  *
- * Copyright (c) 1996-2000 Mike Gleason, NCEMRSoft.
+ * Copyright (c) 1996-2001 Mike Gleason, NCEMRSoft.
  * All rights reserved.
  *
  * A simple, non-interactive utility to send files to a remote FTP server.
@@ -21,10 +21,12 @@
 #	include "..\ncftp\util.h"
 #	include "..\ncftp\spool.h"
 #	include "..\ncftp\pref.h"
+#	include "..\ncftp\getline.h"
 #else
 #	include "../ncftp/util.h"
 #	include "../ncftp/spool.h"
 #	include "../ncftp/pref.h"
+#	include "../ncftp/getline.h"
 #endif
 
 #include "gpshare.h"
@@ -78,12 +80,17 @@ Usage(void)
 	(void) fprintf(fp, "\
   -S XX  Upload into temporary files suffixed by XX.\n\
   -DD    Delete local file after successfully uploading it.\n\
-  -b     Run in background (submit job to \"ncftpbatch\").\n\
+  -b     Run in background (submit job to \"ncftpbatch\" and run).\n\
+  -bb    Same as \"-b\" but queue only (do not run \"ncftpbatch\").\n\
   -E     Use regular (PORT) data connections.\n\
   -F     Use passive (PASV) data connections (default).\n\
-  -y     Try using \"SITE UTIME\" to preserve timestamps on remote host.\n\
+  -y     Try using \"SITE UTIME\" to preserve timestamps on remote host.\n");
+	(void) fprintf(fp, "\
   -B XX  Try setting the SO_SNDBUF size to XX.\n\
   -r XX  Redial XX times until connected.\n\
+  -W XX  Send raw FTP command XX after logging in.\n\
+  -X XX  Send raw FTP command XX after each file transferred.\n\
+  -Y XX  Send raw FTP command XX before logging out.\n\
   -R     Recursive mode; copy whole directory trees.\n");
 	(void) fprintf(fp, "\nExamples:\n\
   ncftpput -u gleason -p my.password Elwood.probe.net /home/gleason stuff.txt\n\
@@ -103,8 +110,9 @@ Usage(void)
 
 
 static void
-Abort(int sigNumUNUSED)
+Abort(int UNUSED(sigNum))
 {
+	LIBNCFTP_USE_VAR(sigNum);
 	signal(SIGINT, Abort);
 
 	/* Hopefully the I/O operation in progress
@@ -125,7 +133,7 @@ Abort(int sigNumUNUSED)
 
 
 static int
-Copy(FTPCIPtr cip, const char *const dstdir, char **const files, const int rflag, const int xtype, const int appendflag, const char *const tmppfx, const char *const tmpsfx, const int resumeflag, const int deleteflag)
+Copy(FTPCIPtr cip, const char *const dstdir, char **const files, const int rflag, const int xtype, const int appendflag, const char *const tmppfx, const char *const tmpsfx, const int resumeflag, const int deleteflag, const char *const perfilecmd)
 {
 	int i;
 	int result;
@@ -146,6 +154,8 @@ Copy(FTPCIPtr cip, const char *const dstdir, char **const files, const int rflag
 		if (result != 0) {
 			FTPPerror(cip, result, kErrCouldNotStartDataTransfer, "ncftpput", file);
 			rc = result;
+		} else {
+			(void) AdditionalCmd(cip, perfilecmd, file);
 		}
 	}
 	return (rc);
@@ -181,7 +191,7 @@ main(int argc, char **argv)
 	char *ufilep;
 	const char *udirp;
 	char ufile[256];
-	char *password;
+	char precmd[128], postcmd[128], perfilecmd[128];
 
 	InitWinsock();
 #ifdef SIGPOLL
@@ -212,8 +222,11 @@ main(int argc, char **argv)
 	fi.errLog = stderr;
 	(void) STRNCPY(fi.user, "anonymous");
 	progmeters = GetDefaultProgressMeterSetting();
+	precmd[0] = '\0';
+	postcmd[0] = '\0';
+	perfilecmd[0] = '\0';
 
-	while ((c = getopt(argc, argv, "P:u:j:p:e:d:U:t:mar:RvVf:AT:S:EFcyZzDbB:")) > 0) switch(c) {
+	while ((c = getopt(argc, argv, "P:u:j:p:e:d:U:t:mar:RvVf:AT:S:EFcyZzDbB:W:X:Y:")) > 0) switch(c) {
 		case 'P':
 			fi.port = atoi(optarg);	
 			break;
@@ -314,6 +327,15 @@ main(int argc, char **argv)
 		case 'B':
 			fi.dataSocketSBufSize = (size_t) atol(optarg);	
 			break;
+		case 'W':
+			STRNCPY(precmd, optarg);
+			break;
+		case 'X':
+			STRNCPY(perfilecmd, optarg);
+			break;
+		case 'Y':
+			STRNCPY(postcmd, optarg);
+			break;
 		default:
 			Usage();
 	}
@@ -344,12 +366,7 @@ main(int argc, char **argv)
 
 	if (strcmp(fi.user, "anonymous") && strcmp(fi.user, "ftp")) {
 		if (fi.pass[0] == '\0') {
-			password = GetPass2("Password: ");		
-			if (password != NULL) {
-				(void) STRNCPY(fi.pass, password);
-				/* Don't leave cleartext password in memory. */
-				(void) memset(password, 0, strlen(fi.pass));
-			}
+			(void) gl_getpass("Password: ", fi.pass, sizeof(fi.pass));
 		}
 	}
 
@@ -396,6 +413,9 @@ main(int argc, char **argv)
 				rflag,
 				deleteflag,
 				fi.dataPortMode,
+				precmd,
+				perfilecmd,
+				postcmd,
 				(time_t) 0	/* when: now */
 			);
 			if (result == 0) {
@@ -404,7 +424,9 @@ main(int argc, char **argv)
 			}
 		}
 		if (spooled > 0) {
-			RunBatch(0, NULL);
+			if (batchmode == 1) {
+				RunBatch(0, NULL);
+			}
 			DisposeWinsock(0);
 			exit(kExitSuccess);
 		}
@@ -431,9 +453,12 @@ main(int argc, char **argv)
 			FTPPerror(&fi, result, kErrUmaskFailed, "ncftpput", "could not set umask");
 	}
 
+	errstr = "could not run pre-command remote host";
+	(void) AdditionalCmd(&fi, precmd, NULL);
+
 	if (dstdir != NULL) {
 		es = kExitChdirTimedOut;
-		if (wantMkdir != 0)  {
+		if (wantMkdir != 0) {
 			errstr = "could not create and chdir on remote host";
 			result = FTPChdir3(&fi, dstdir, NULL, 0, kChdirOneSubdirAtATime|kChdirAndMkdir);
 		} else {
@@ -455,7 +480,7 @@ main(int argc, char **argv)
 		es = kExitXferTimedOut;
 		(void) signal(SIGINT, Abort);
 		if (ftpcat == 0) {
-			if (Copy(&fi, "", files, rflag, xtype, appendflag, (const char *) tmppfx, (const char *) tmpsfx, resumeflag, deleteflag) < 0)
+			if (Copy(&fi, "", files, rflag, xtype, appendflag, (const char *) tmppfx, (const char *) tmpsfx, resumeflag, deleteflag, perfilecmd) < 0)
 				es = kExitXferFailed;
 			else
 				es = kExitSuccess;
@@ -463,10 +488,15 @@ main(int argc, char **argv)
 			fi.progress = (FTPProgressMeterProc) 0;
 			if (FTPPutOneFile2(&fi, NULL, argv[optind + 1], xtype, STDIN_FILENO, appendflag, tmppfx, tmpsfx) < 0)
 				es = kExitXferFailed;
-			else
+			else {
 				es = kExitSuccess;
+				(void) AdditionalCmd(&fi, perfilecmd, argv[optind + 1]);
+			}
 		}
 	}
+
+	errstr = "could not run post-command remote host";
+	(void) AdditionalCmd(&fi, postcmd, NULL);
 	
 	(void) FTPCloseHost(&fi);
 	DisposeWinsock(0);
